@@ -14,7 +14,6 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 
-#include <linux/fs.h>
 #include <linux/ext2_fs.h>
 
 #include "ext2fs.h"
@@ -31,6 +30,8 @@ errcode_t ext2fs_read_bb_FILE(ext2_filsys fs, FILE *f,
 	int		count;
 	char		buf[128];
 
+	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);
+
 	if (!*bb_list) {
 		retval = badblocks_list_create(bb_list, 10);
 		if (retval)
@@ -40,7 +41,7 @@ errcode_t ext2fs_read_bb_FILE(ext2_filsys fs, FILE *f,
 	while (!feof (f)) {
 		if (fgets(buf, sizeof(buf), f) == NULL)
 			break;
-		count = sscanf(buf, "%lu", &blockno);
+		count = sscanf(buf, "%u", &blockno);
 		if (count <= 0)
 			continue;
 		if ((blockno < fs->super->s_first_data_block) ||
