@@ -3,15 +3,14 @@
  *
  * For copyright information, see copyright.h.
  */
+
+#ifdef HAS_STDLIB_H
+#include <stdlib.h>
+#endif
 #include "ss_internal.h"
 #include "copyright.h"
 #define	size	sizeof(ss_data *)
 
-#ifndef lint
-static char const rcsid[] =
-    "$Header: /mit/krb5/.cvsroot/src/util/ss/invocation.c,v 1.1 1993/06/03 12:31:05 tytso Exp $";
-#endif
-
 int ss_create_invocation(subsystem_name, version_string, info_ptr,
 			 request_table_ptr, code_ptr)
 	char *subsystem_name, *version_string;
