@@ -16,12 +16,18 @@
  * 94/02/27	- Integrated in Ted's distribution
  */
 
+#include <sys/types.h>
 #include <dirent.h>
 #include <fcntl.h>
+#ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#endif
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
+#ifdef HAVE_ERRNO_H
+#include <errno.h>
+#endif
 #include <sys/param.h>
 #include <sys/stat.h>
 #include <linux/ext2_fs.h>
@@ -53,7 +59,7 @@ static void volatile fatal_error (const char * fmt_string, int errcode)
 	exit (errcode);
 }
 
-#define usage() fatal_error ("usage: %s [-RV] [-+=csu] [-v version] files...\n", \
+#define usage() fatal_error ("usage: %s [-RV] [-+=acdisSu] [-v version] files...\n", \
 			     1)
 
 static int decode_arg (int * i, int argc, char ** argv)
@@ -77,10 +83,28 @@ static int decode_arg (int * i, int argc, char ** argv)
 			case 'V':
 				verbose = 1;
 				break;
+#ifdef	EXT2_APPEND_FL
+			case 'a':
+				rf |= EXT2_APPEND_FL;
+				rem = 1;
+				break;
+#endif
 			case 'c':
 				rf |= EXT2_COMPR_FL;
 				rem = 1;
 				break;
+#ifdef	EXT2_NODUMP_FL
+			case 'd':
+				rf |= EXT2_NODUMP_FL;
+				rem = 1;
+				break;
+#endif
+#ifdef	EXT2_IMMUTABLE_FL
+			case 'i':
+				rf |= EXT2_IMMUTABLE_FL;
+				rem = 1;
+				break;
+#endif
 			case 's':
 				rf |= EXT2_SECRM_FL;
 				rem = 1;
@@ -116,9 +140,24 @@ static int decode_arg (int * i, int argc, char ** argv)
 			case 'S':
 				af |= EXT2_SYNC_FL;
 				break;
+#ifdef	EXT2_APPEND_FL
+			case 'a':
+				af |= EXT2_APPEND_FL;
+				break;
+#endif
 			case 'c':
 				af |= EXT2_COMPR_FL;
 				break;
+#ifdef	EXT2_NODUMP_FL
+			case 'd':
+				af |= EXT2_NODUMP_FL;
+				break;
+#endif
+#ifdef	EXT2_IMMUTABLE_FL
+			case 'i':
+				af |= EXT2_IMMUTABLE_FL;
+				break;
+#endif
 			case 's':
 				af |= EXT2_SECRM_FL;
 				break;
@@ -137,9 +176,24 @@ static int decode_arg (int * i, int argc, char ** argv)
 			case 'S':
 				sf |= EXT2_SYNC_FL;
 				break;
+#ifdef	EXT2_APPEND_FL
+			case 'a':
+				sf |= EXT2_APPEND_FL;
+				break;
+#endif
 			case 'c':
 				sf |= EXT2_COMPR_FL;
 				break;
+#ifdef	EXT2_NODUMP_FL
+			case 'd':
+				sf |= EXT2_NODUMP_FL;
+				break;
+#endif
+#ifdef	EXT2_IMMUTABLE_FL
+			case 'i':
+				sf |= EXT2_IMMUTABLE_FL;
+				break;
+#endif
 			case 's':
 				sf |= EXT2_SECRM_FL;
 				break;
@@ -174,7 +228,7 @@ static void change_attributes (const char * name)
 		if (verbose)
 		{
 			printf ("Flags of %s set as ", name);
-			print_flags (stdout, sf);
+			print_flags (stdout, sf, 0);
 			printf ("\n");
 		}
 		if (fsetflags (name, sf) == -1)
@@ -194,7 +248,7 @@ static void change_attributes (const char * name)
 			if (verbose)
 			{
 				printf ("Flags of %s set as ", name);
-				print_flags (stdout, flags);
+				print_flags (stdout, flags, 0);
 				printf ("\n");
 			}
 			if (fsetflags (name, flags) == -1)
@@ -216,12 +270,17 @@ static void change_attributes (const char * name)
 
 static int chattr_dir_proc (const char * dir_name, struct dirent * de, void * private)
 {
-	char path[MAXPATHLEN];
-
 	if (strcmp (de->d_name, ".") && strcmp (de->d_name, ".."))
 	{
+	        char *path;
+
+		path = malloc(strlen (dir_name) + 1 + strlen (de->d_name) + 1);
+		if (!path)
+			fatal_error("Couldn't allocate path variable "
+				    "in chattr_dir_proc", 1);
 		sprintf (path, "%s/%s", dir_name, de->d_name);
 		change_attributes (path);
+		free(path);
 	}
 	return 0;
 }
